#!/usr/bin/env python3
# coding: utf-8

from shutil import rmtree
import os
import stat
import subprocess
import argparse as ap
from textwrap import dedent
import zipfile

AMS_INSTALLER_VER = "1.2"

DEFAULT_INSTALL_SCRIPT_NAME = "install.sh"
DEFAULT_PACKAGES = [
    "./AMS_INSTALL.zip",
    "/home/ams/AMS_INSTALL.zip",
    "/home/pi/AMS_INSTALL.zip",
    "/AMS_INSTALL.zip"
]

def find_all_files_in_path(name, path):
    result = []
    for root, dirs, files in os.walk(path):
        if name in files:
            result.append(os.path.join(root, name))
    return result

def logit(output):
    print("ams_installer: {}".format(output), flush=True)

if __name__ == '__main__':
    parser = ap.ArgumentParser(
        formatter_class=ap.RawTextHelpFormatter,
        description=dedent('''
        The AMS installer unzips packages and runs installation scripts.
        The primary use of this tool is for automated installation of packages
        when run at boot.

        Note: Currently only zip files are supported
        '''),
        epilog=dedent(
        '''
        Examples:

            ./ams_installer.py

                    Search for AMS_INSTALL.zip, unzip it and search and run
                    'install.sh', if present


            ./ams_installer.py /home/ams/package.zip

                    Unzip package.zip and search and run 'install.sh', if present

        '''))
    parser.add_argument('-n', '--no_reboot', action="store_true", help=dedent('''\
                        Do not reboot after installing the package. Default is
                        to reboot after executing 'install.sh'.
                        '''))
    parser.add_argument('-V', '--version', action='store_true', help=dedent('''\
                        Print the version of this tool and exit.
                        '''))
    parser.add_argument('packages', nargs='*', default=DEFAULT_PACKAGES, help=dedent('''\
                        Package to decompress. The tool searches for an
                        executable script in the extracted files called
                        'install.sh' and runs it if available.

                        If no package is supplied the tool searches for a package
                        called AMS_INSTALL.zip in this location order:
                        {}
                        and attempts to extract and install it.

                        During installation the package is renamed to
                        '<package>.IN_PROGRESS'.

                        After installation the package is renamed to
                        '<package>.SUCCESS' or '<package>.FAIL' depending
                        on the return code of the 'install.sh'

                        The system is then rebooted.
                        '''.format(DEFAULT_PACKAGES)))

    args = parser.parse_args()

    logit("version - {}".format(AMS_INSTALLER_VER))
    if args.version:
        exit(0)

    packages = [os.path.abspath(x) for x in args.packages if zipfile.is_zipfile(x)]
    # convert through set to eliminate duplicates
    packages = list(set(packages))

    if not len(packages):
        logit("No packages to install. Exiting.")
        exit(0)

    logit("package(s) to install:\n\t'{}'".format("\n\t".join(packages)))

    # start installing packages one by one
    for base_package in packages:

        install_good = False

        base_package_name = os.path.basename(base_package)
        package = base_package + ".IN_PROGRESS"
        os.rename(base_package, package)

        # create temporary directory for extraction
        temp_dir_cmd = ['mktemp', '-d', '-t', "{}-XXXXX".format(base_package_name)]
        result = subprocess.run(temp_dir_cmd, stdout=subprocess.PIPE)
        if result.returncode != 0:
            raise OSError("Error, creating temp dir for package '{}'".format(base_package_name))
        temp_dir = result.stdout.decode('utf-8').replace('\n', '')
        logit("Temporary dir for extraction: '{}'".format(temp_dir))

        # extract the package
        zip_ref = zipfile.ZipFile(package, 'r')
        zip_ref.extractall(temp_dir)
        zip_ref.close()

        # find the install script
        install_scripts = find_all_files_in_path(DEFAULT_INSTALL_SCRIPT_NAME, temp_dir)
        if len(install_scripts) > 1:
            logit("Found multiple install scripts: '{}'".format(install_scripts))
        elif len(install_scripts) == 0:
            logit("No '{}' found. Marking as failed")
            install_good = False
        else:
            install_script = install_scripts[0]
            if not os.path.isfile(install_script):
                logit("'{}' is not a file.".format(install_script))
            logit("Running install script: '{}'".format(install_script))

            # make sure install script is executable
            os.chmod(install_script, os.stat(install_script).st_mode | stat.S_IXOTH)
            os.chdir(os.path.dirname(install_script))

            # run the install script
            install_cmd = [install_script]
            result = subprocess.run(install_cmd)
            if result.returncode != 0:
                install_good = False
            else:
                install_good = True

        # mark package as Success/Fail
        if install_good:
            suffix = ".SUCCESS"
            logit("package '{}' install success.".format(base_package_name))
        else:
            suffix = ".FAIL"
            logit("package '{}' install fail.".format(base_package_name))
        os.rename(package, base_package + suffix)

        # delete temp directory
        rmtree(temp_dir, ignore_errors=True)

    # make sure filesystem is sync'd
    subprocess.run(['sync'])
    logit("installation complete.")

    if args.no_reboot:
        exit(0)
    else:
        # reboot the system, we are done
        logit("rebooting ... now.")
        subprocess.run(['reboot', '-n'])
